#include "Cv10_main.h"

//---------------------
//----    SCENE    ----
//---------------------

/// Reloads all shaders
void reload_shaders()
{
	// Program for objects without textures and without lighting (for the light object)
	nolit_program.Init();
	nolit_program.AddVertexShader("Shaders/nolit_vertex.glsl");
	nolit_program.AddFragmentShader("Shaders/nolit_fragment.glsl");
	nolit_program.Link();

	BlinnPhongOriginal_program.Init();
	BlinnPhongOriginal_program.AddVertexShader("Shaders/texture_vertex.glsl");
	BlinnPhongOriginal_program.AddFragmentShader("Shaders/BlinnPhongOriginal_fragment.glsl");
	BlinnPhongOriginal_program.Link();

	BlinnPhongNormalized_program.Init();
	BlinnPhongNormalized_program.AddVertexShader("Shaders/texture_vertex.glsl");
	BlinnPhongNormalized_program.AddFragmentShader("Shaders/BlinnPhongNormalized_fragment.glsl");
	BlinnPhongNormalized_program.Link();

	CookTorrance_program.Init();
	CookTorrance_program.AddVertexShader("Shaders/texture_vertex.glsl");
	CookTorrance_program.AddFragmentShader("Shaders/CookTorrance_fragment.glsl");
	CookTorrance_program.Link();

	OrenNayar_program.Init();
	OrenNayar_program.AddVertexShader("Shaders/texture_vertex.glsl");
	OrenNayar_program.AddFragmentShader("Shaders/OrenNayar_fragment.glsl");
	OrenNayar_program.Link();

	AshikhminShirley_program.Init();
	AshikhminShirley_program.AddVertexShader("Shaders/texture_vertex.glsl");
	AshikhminShirley_program.AddFragmentShader("Shaders/AshikhminShirley_fragment.glsl");
	AshikhminShirley_program.Link();

	cout << "Shaders are reloaded" << endl;
}

/// Initializes all objects of our scene
void init_scene()
{
	//----------------------------------------------
	//--  Display some information about the OpenGL

	cout << "Vendor   : " << (const char *)glGetString(GL_VENDOR) << endl;
	cout << "Renderer : " << (const char *)glGetString(GL_RENDERER) << endl;
	cout << "Version  : " << (const char *)glGetString(GL_VERSION) << endl;

	cout << endl;

	//----------------------------------------------
	//--  Load shaders

	reload_shaders();
	
	//----------------------------------------------
	//--  Prepare materials

	WhiteMaterial_ubo.Init();
	WhiteMaterial_ubo.SetMaterial(PhongMaterial::CreateBasicMaterial(glm::vec3(1.0f, 1.0f, 1.0f), true, 200.0f, 1.0f));
	WhiteMaterial_ubo.UpdateOpenGLData();

	//----------------------------------------------
	//--  Prepare geometries
	
	geom_cube = CreateCube();
	geom_sphere = CreateSphere();
	geom_torus = CreateTorus();
	geom_cylinder = CreateCylinder();
	geom_capsule = CreateCapsule();
	geom_teapot = CreateTeapot();

	//----------------------------------------------
	//--  Positions of objects in the scene

	// Light's sphere model matrix (its position)
	LightModel_ubo.Init();
	LightModel_ubo.SetMatrix(glm::mat4(1.0f));
	LightModel_ubo.UpdateOpenGLData();

	// Central object's model matrix
	CentralObjectModel_ubo.Init();
	CentralObjectModel_ubo.SetMatrix(glm::mat4(1.0f));
	CentralObjectModel_ubo.UpdateOpenGLData();

	//----------------------------------------------
	//--  Prepare all cameras
	
	// Create the basic camera through which we look into the scene
	the_camera.SetCenter(glm::vec3(0.0f));
	the_camera.SetCamera(-0.5f, -0.5f, 15.0f);
	CameraData_ubo.Init();

	//----------------------------------------------
	//--  Prepare the lights

	// Create only a single directional light which direction corresponds to the position of the sun in the cubemap
	PhongLights_ubo.Init();
	PhongLights_ubo.SetGlobalAmbient(glm::vec3(0.2f));
	PhongLights_ubo.PhongLights.clear();
	PhongLights_ubo.PhongLights.push_back(
		PhongLight::CreateDirectionalLight(
		glm::vec3(1.0f, 1.0f, 1.0f),
		glm::vec3(0.0f), glm::vec3(1.0f), glm::vec3(1.0f)));
	PhongLights_ubo.UpdateOpenGLData();
}

/// Updates the scene: performs animations, updates the data of the buffers, etc.
void update_scene(int app_time_diff_ms)
{
	// Update the data of the preset material
	switch (material_preset)
	{
	case MATERIAL_PRESET_USER:			break;		// Do nothing
	case MATERIAL_PRESET_PLASTIC_USER:	material_fresnel = glm::vec3(0.04f, 0.04f, 0.04f);		break;
	case MATERIAL_PRESET_PLASTIC_RED:	material_fresnel = glm::vec3(0.04f, 0.04f, 0.04f);		material_diffuse = glm::vec3(1.0f, 0.0f, 0.0f);	break;
	case MATERIAL_PRESET_PLASTIC_GREEN:	material_fresnel = glm::vec3(0.04f, 0.04f, 0.04f);		material_diffuse = glm::vec3(0.0f, 1.0f, 0.0f);	break;
	case MATERIAL_PRESET_PLASTIC_BLUE:	material_fresnel = glm::vec3(0.04f, 0.04f, 0.04f);		material_diffuse = glm::vec3(0.0f, 0.0f, 1.0f);	break;
	case MATERIAL_PRESET_IRON:			material_fresnel = glm::vec3(0.56f, 0.57f, 0.58f);		material_diffuse = glm::vec3(0.0f, 0.0f, 0.0f);	break;
	case MATERIAL_PRESET_COPPER:		material_fresnel = glm::vec3(0.95f, 0.64f, 0.54f);		material_diffuse = glm::vec3(0.0f, 0.0f, 0.0f);	break;
	case MATERIAL_PRESET_GOLD:			material_fresnel = glm::vec3(1.00f, 0.71f, 0.29f);		material_diffuse = glm::vec3(0.0f, 0.0f, 0.0f);	break;
	case MATERIAL_PRESET_ALUMINIUM:		material_fresnel = glm::vec3(0.91f, 0.92f, 0.92f);		material_diffuse = glm::vec3(0.0f, 0.0f, 0.0f);	break;
	case MATERIAL_PRESET_SILVER:		material_fresnel = glm::vec3(0.95f, 0.93f, 0.88f);		material_diffuse = glm::vec3(0.0f, 0.0f, 0.0f);	break;
	default: break;
	}

	//----------------------------------------------
	//--  Data of the lights

	const float light_distance = 10.0f;
	glm::vec3 light_position = light_distance * glm::vec3(
		cosf(light_pos / 10.0f) * sinf(light_pos),
		sinf(light_pos / 10.0f),
		cosf(light_pos / 10.0f) * cosf(light_pos));

	PhongLights_ubo.PhongLights[0] = PhongLight::CreatePointLight(
		light_position,
		glm::vec3(ambient_light_intensity),
		glm::vec3(light_intensity),
		glm::vec3(light_intensity));
	PhongLights_ubo.UpdateOpenGLData();

	LightModel_ubo.SetMatrix(
		glm::translate(glm::mat4(1.0f), light_position) * 
		glm::scale(glm::mat4(1.0f), glm::vec3(0.1f)));
	LightModel_ubo.UpdateOpenGLData();

	//----------------------------------------------
	//--  Data of the main camera
	
	CameraData_ubo.SetProjection(
		glm::perspective(glm::radians(45.0f), float(win_width) / float(win_height), 0.1f, 1000.0f));
	CameraData_ubo.SetCamera(the_camera);
	CameraData_ubo.UpdateOpenGLData();
}

/// Renders the whole frame
void render_scene()
{
	// Render into the main window
	glBindFramebuffer(GL_FRAMEBUFFER, 0);
	glViewport(0, 0, win_width, win_height);

	// Clear the framebuffer, clear the color and depth
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glClearDepth(1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glEnable(GL_DEPTH_TEST);

	// Set the data of the camera and the lights
	CameraData_ubo.BindBuffer(DEFAULT_CAMERA_BINDING);
	PhongLights_ubo.BindBuffer(DEFAULT_LIGHTS_BINDING);

	bool can_render = false;
	if ((material_type == MATERIAL_TYPE_BLINN_PHONG_ORIGINAL) &&
		BlinnPhongOriginal_program.IsValid())
	{
		BlinnPhongOriginal_program.Use();
		BlinnPhongOriginal_program.Uniform3fv("material_diffuse", 1, glm::value_ptr(material_diffuse));
		BlinnPhongOriginal_program.Uniform3fv("material_specular", 1, glm::value_ptr(material_specular));
		BlinnPhongOriginal_program.Uniform1f("material_shininess", material_shininess);
		can_render = true;
	}
	else if ((material_type == MATERIAL_TYPE_BLINN_PHONG_NORMALIZED) &&
		BlinnPhongNormalized_program.IsValid())
	{
		BlinnPhongNormalized_program.Use();
		BlinnPhongNormalized_program.Uniform3fv("material_diffuse", 1, glm::value_ptr(material_diffuse));
		BlinnPhongNormalized_program.Uniform3fv("material_specular", 1, glm::value_ptr(material_specular));
		BlinnPhongNormalized_program.Uniform1f("material_shininess", material_shininess);
		can_render = true;
	}
	else if ((material_type == MATERIAL_TYPE_COOK_TORRANCE) &&
		CookTorrance_program.IsValid())
	{
		CookTorrance_program.Use();
		CookTorrance_program.Uniform3fv("material_fresnel", 1, glm::value_ptr(material_fresnel));
		CookTorrance_program.Uniform3fv("material_diffuse", 1, glm::value_ptr(material_diffuse));
		CookTorrance_program.Uniform1f("material_roughness", material_roughness);
		can_render = true;
	}
	else if ((material_type == MATERIAL_TYPE_OREN_NAYAR) &&
		OrenNayar_program.IsValid())
	{
		OrenNayar_program.Use();
		OrenNayar_program.Uniform3fv("material_diffuse", 1, glm::value_ptr(material_diffuse));
		OrenNayar_program.Uniform1f("material_roughness", material_roughness);
		can_render = true;
	}
	else if ((material_type == MATERIAL_TYPE_ASHIKHMIN_SHIRLEY) &&
		AshikhminShirley_program.IsValid())
	{
		AshikhminShirley_program.Use();
		AshikhminShirley_program.Uniform3fv("material_fresnel", 1, glm::value_ptr(material_fresnel));
		AshikhminShirley_program.Uniform3fv("material_diffuse", 1, glm::value_ptr(material_diffuse));
		AshikhminShirley_program.Uniform1f("material_shininess_tangent", material_shininess_tangent);
		AshikhminShirley_program.Uniform1f("material_shininess_bitangent", material_shininess_bitangent);
		can_render = true;
	}

	if (can_render)
	{
		// Set object's position
		CentralObjectModel_ubo.BindBuffer(DEFAULT_OBJECT_BINDING);

		// Render proper object
		switch (central_object)
		{
		case CENTRAL_OBJECT_CUBE:		geom_cube.BindVAO();		geom_cube.Draw();		break;
		case CENTRAL_OBJECT_SPHERE:		geom_sphere.BindVAO();		geom_sphere.Draw();		break;
		case CENTRAL_OBJECT_TORUS:		geom_torus.BindVAO();		geom_torus.Draw();		break;
		case CENTRAL_OBJECT_CYLINDER:	geom_cylinder.BindVAO();	geom_cylinder.Draw();	break;
		case CENTRAL_OBJECT_CAPSULE:	geom_capsule.BindVAO();		geom_capsule.Draw();	break;
		case CENTRAL_OBJECT_TEAPOT:		geom_teapot.BindVAO();		geom_teapot.Draw();		break;
		}
	}

	// Render the light's sphere
	if (nolit_program.IsValid())
	{
		nolit_program.Use();

		LightModel_ubo.BindBuffer(DEFAULT_OBJECT_BINDING);
		WhiteMaterial_ubo.BindBuffer(DEFAULT_MATERIAL_BINDING);
		geom_sphere.BindVAO();
		geom_sphere.Draw();
	}

	// Reset the VAO and the program
	glBindVertexArray(0);
	glUseProgram(0);
	glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

//-------------------
//----    GUI    ----
//-------------------

/// Callback when user clicks on Reload button
void TW_CALL reload(void *)
{
	reload_shaders();
}

void init_gui()
{
	central_object = CENTRAL_OBJECT_SPHERE;

	light_pos = 0.0f;
	light_intensity = 1.0f;
	ambient_light_intensity = 0.01f;

	material_preset = MATERIAL_PRESET_USER;
	material_diffuse = glm::vec3(1.0f, 0.0f, 0.0f);
	material_fresnel = glm::vec3(0.04, 0.04, 0.04);
	material_specular = glm::vec3(1.0, 1.0, 1.0);
	material_roughness = 0.5f;
	material_shininess = 10.0f;
	material_shininess_tangent = material_shininess;
	material_shininess_bitangent = material_shininess;

	material_type = MATERIAL_TYPE_BLINN_PHONG_ORIGINAL;

	// Initialize AntTweakBar
	TwInit(TW_OPENGL_CORE, nullptr);
	TwWindowSize(win_width, win_height);
	TwGLUTModifiersFunc(glutGetModifiers);

	// Initialize GUI
	the_gui = TwNewBar("Parameters");
	TwDefine("Parameters size='270 400'");
	TwDefine("Parameters valueswidth=100");
	TwAddButton(the_gui, "Reload", reload, nullptr, nullptr);

	// Parameters of the light
	TwAddSeparator(the_gui, nullptr, nullptr);
	TwAddVarRW(the_gui, "Light direction", TW_TYPE_FLOAT, &light_pos, "min=-15.0 max=15.0 step=0.03");
	TwAddVarRW(the_gui, "Light intensity", TW_TYPE_FLOAT, &light_intensity, "min=0.0 max=2.0 step=0.01");
	TwAddVarRW(the_gui, "Ambient intensity", TW_TYPE_FLOAT, &ambient_light_intensity, "min=0.0 max=0.1 step=0.01");

	// Parameters of the material
	TwAddSeparator(the_gui, nullptr, nullptr);
	TwEnumVal material_preset_type[10] = {
		{MATERIAL_PRESET_USER,			"User defined"},
		{MATERIAL_PRESET_PLASTIC_USER,	"Plastic, user color"},
		{MATERIAL_PRESET_PLASTIC_RED,	"Plastic red"},
		{MATERIAL_PRESET_PLASTIC_GREEN,	"Plastic green"},
		{MATERIAL_PRESET_PLASTIC_BLUE,	"Plastic blue"},
		{MATERIAL_PRESET_IRON,		"Iron"},
		{MATERIAL_PRESET_COPPER,	"Copper"},
		{MATERIAL_PRESET_GOLD,		"Gold"},
		{MATERIAL_PRESET_ALUMINIUM,	"Aluminium"},
		{MATERIAL_PRESET_SILVER,	"Silver"},
	};
	TwType materialPresetType = TwDefineEnum("MaterialPresetType", material_preset_type, 10);
	TwAddVarRW(the_gui, "Material Preset", materialPresetType, &material_preset, nullptr);

	TwAddVarRW(the_gui, "Diffuse color", TW_TYPE_COLOR3F, &material_diffuse, nullptr);
	TwAddVarRW(the_gui, "Fresnel color", TW_TYPE_COLOR3F, &material_fresnel, nullptr);
	TwAddVarRW(the_gui, "Specular color", TW_TYPE_COLOR3F, &material_specular, nullptr);
	TwAddVarRW(the_gui, "Roughness", TW_TYPE_FLOAT, &material_roughness, "min=0.0 max=5.0 step=0.01");
	TwAddVarRW(the_gui, "Shininess", TW_TYPE_FLOAT, &material_shininess, "min=0.0 max=500.0 step=1.0");
	TwAddVarRW(the_gui, "Shininess (tangent)", TW_TYPE_FLOAT, &material_shininess_tangent, "min=0.0 max=500.0 step=1.0");
	TwAddVarRW(the_gui, "Shininess (bitangent)", TW_TYPE_FLOAT, &material_shininess_bitangent, "min=0.0 max=500.0 step=1.0");

	TwEnumVal material_types[5] = {
		{MATERIAL_TYPE_BLINN_PHONG_ORIGINAL, "Blinn-Phong Original"},
		{MATERIAL_TYPE_BLINN_PHONG_NORMALIZED, "Blinn-Phong Normalized"},
		{MATERIAL_TYPE_COOK_TORRANCE, "Cook-Torrance"},
		{MATERIAL_TYPE_OREN_NAYAR, "Oren-Nayar"},
		{MATERIAL_TYPE_ASHIKHMIN_SHIRLEY, "Ashikhmin-Shirley"},
	};
	TwType materialTypeType = TwDefineEnum("MaterialTypes", material_types, 5);
	TwAddVarRW(the_gui, "Material", materialTypeType, &material_type, nullptr);

	// Other parameters
	TwAddSeparator(the_gui, nullptr, nullptr);

	// Central object
	TwEnumVal objects[6] = {
		{CENTRAL_OBJECT_CUBE, "Cube"},
		{CENTRAL_OBJECT_SPHERE, "Sphere"},
		{CENTRAL_OBJECT_TORUS, "Torus"},
		{CENTRAL_OBJECT_CYLINDER, "Cylinder"},
		{CENTRAL_OBJECT_CAPSULE, "Capsule"},
		{CENTRAL_OBJECT_TEAPOT, "Teapot"},
	};
	TwType objectType = TwDefineEnum("Objects", objects, 6);
	TwAddVarRW(the_gui, "Central object", objectType, &central_object, nullptr);
}

//---------------------------
//----    APPLICATION    ----
//---------------------------

/// GLUT callback - when the user presses a key. See glutKeyboardFunc for more info
void on_keyboard_func(unsigned char key, int x, int y)
{
	// Inform AntTweakBar
	if (TwEventKeyboardGLUT(key, x, y))
		return;		// Already handled by AntTweakBar

	switch (key)
	{
	case 27:		// Escape key
		exit(0);
		break;
	}
}

/// GLUT callback - when the user presses a special key (like F1 key). See glutSpecialFunc for more info
void on_special_func(int key, int x, int y)
{
	// Inform AntTweakBar
	if (TwEventSpecialGLUT(key, x, y))
		return;		// Already handled by AntTweakBar
}

/// GLUT callback - when the user presses or releases a mouse button. See glutMouseFunc for more info
void on_mouse_func(int button, int state, int x, int y)
{
	// Inform AntTweakBar
	if (TwEventMouseButtonGLUT(button, state, x, y))
		return;		// Already handled by AntTweakBar

	the_camera.OnMouseFunc(button, state, x, y);
}

/// GLUT callback - when the user moves the mouse when holding a botton. See glutMotionFunc for more info
void on_motion_func(int x, int y)
{
	// Inform AntTweakBar
	if (TwEventMouseMotionGLUT(x, y))
		return;		// Already handled by AntTweakBar

	the_camera.OnMotionFunc(x, y);
}

/// GLUT callback - when the user moves the mouse when not holding a botton. See glutPassiveMotionFunc for more info
void on_passive_motion_func(int x, int y)
{
	// Inform AntTweakBar
	if (TwEventMouseMotionGLUT(x, y))
		return;		// Already handled by AntTweakBar
}

/// GLUT callback - when the window is resized. See glutReshapeFunc for more info
void on_reshape(int width, int height)
{
	win_width = width;
	win_height = height;
	
	glViewport(0, 0, win_width, win_height);
	
	// Inform AntTweakBar
	TwWindowSize(win_width, win_height);
}

/// GLUT callback - when our animation timer requests a new frame. See glutTimerFunc for more info
void on_timer(int)
{
	// Register the timer to be called again in 20 ms, and rerender the scene
	glutTimerFunc(20, on_timer, 0);
	glutPostRedisplay();
}

/// GLUT callback - when the window needs to be redrawn. See glutDisplayFunc for more info
void on_display()
{
	//--  Update all the data

	// Update the application time
	int current_glut_time = glutGet(GLUT_ELAPSED_TIME);
	int app_time_diff_ms = current_glut_time - last_glut_time;
	app_time_ms += app_time_diff_ms;
	last_glut_time = current_glut_time;

	// Update the scene
	update_scene(app_time_diff_ms);

	//--  Render the scene

	// Render the scene
	render_scene();
	
	// Render the GUI
	TwDraw();

	// Swaps the front and back buffer (double-buffering)
	glutSwapBuffers();
}

/// Callback function to be called when we make an error in OpenGL
void GLAPIENTRY simple_debug_callback(GLenum source, GLenum type, GLuint id, GLenum severity, GLsizei length, const char* message, const void* userParam)
{
	switch (type)			// Report only some errors
	{
	case GL_DEBUG_TYPE_ERROR:
	case GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR:
	case GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR:
		cout << message << endl;		// Put the breakpoint here
		return;
	default:
		return;
	}
}

/// C main function :-)
int main(int argc, char ** argv)
{
	// Initialize GLUT
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
	glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);		// Return from the main loop

	// Set OpenGL Context parameters
	glutInitContextVersion(4, 3);				// Use OpenGL 4.3
	glutInitContextProfile(GLUT_CORE_PROFILE);	// Use OpenGL core profile
	glutInitContextFlags(GLUT_DEBUG);			// Use OpenGL debug context

	// Initialize window
	glutInitWindowSize(win_width, win_height);
	glutCreateWindow("PV227 - Cv10: Microfacets");

	// Initialize GLEW
	glewExperimental = GL_TRUE;
	glewInit();

	// Initialize DevIL library
	InitDevIL();

	// Set the OpenGL debug callback
	SetDebugCallback(simple_debug_callback);

	// Register GLUT callbacks
	glutDisplayFunc(on_display);
	glutReshapeFunc(on_reshape);
	glutTimerFunc(20, on_timer, 0);

	glutKeyboardFunc(on_keyboard_func);
	glutSpecialFunc(on_special_func);
	glutMouseFunc(on_mouse_func);
	glutMotionFunc(on_motion_func);
	glutPassiveMotionFunc(on_passive_motion_func);

	// Initialize OpenGL stuff
	init_gui();
	init_scene();

	// Everything is loaded and prepared, reset the application time
	last_glut_time = glutGet(GLUT_ELAPSED_TIME);
	app_time_ms = 0;

	// Run the main loop
	glutMainLoop();

	// Unload AntTweakBar
	TwTerminate();

	return 0;
}